from builtin.stdio import *

# > sort 5
ENGINE_NAME = "ATEngine"


class Scheduler:  # > remove
    know = UsrObject()  # > remove
    now: int = 0  # > remove

    def handle(self):  # > remove
        pass  # > remove


class Updatable:
    __kind__ = "updatable"
    """
    这是一个用于构造可更新对象的基类 | This is a base class for constructing updatable objects
    """
    KNOW = UsrObject()  # 用于记录重要全局数据 | Used to record important global data
    UPDATES = []  # 用于记录所有Updatable的子类(只考虑Src内定义的) |
    #   Used to record subclasses of all Updatable(only considering those defined in Src)

    def __init__(self, _enable: bool = None):
        self._alive = True  # 是否存活. 一旦被置位为False，将不再被更新 | Whether alive. Once set to False, it will no longer be updated
        self._inited = False  # 是否已经初始化 | Whether initialized
        self._enabled = True if _enable is None else _enable  # 被置位时，才允许更新事件 | When set, update events are allowed
        self.know = Updatable.KNOW  # 用于记录重要全局数据 | Used to record important global data
        self.loading()

        if self.__class__ not in Updatable.UPDATES:
            # > insert _Engine_Scheduler.add(self)
            pass

    def loading(self):
        pass

    def init(self, k: UsrObject) -> None:
        """
        * 首次执行step函数前，执行此函数 | called before executing the step function for the first time
        Args:
            k: UsrObject  # 用于记录重要全局数据 | used to record important global data

        Returns:
            None
        """
        pass

    def step(self, k: UsrObject) -> None:
        """
        * 由handle函数调用，执行此函数 | called by the handle function to execute this function
        Args:
            k: UsrObject  # 用于记录重要全局数据 | used to record important global data

        Returns:
            None
        """
        pass

    def halt(self, k: UsrObject) -> None:
        """
        * 由外部停止此对象时调用此 | called when this object is stopped externally
        Args:
            k: UsrObject  # 用于记录重要全局数据 | used to record important global data

        Returns:
            None
        """
        pass

    # ==================================================================================
    def disable(self) -> None:
        """
        * 由用户调用，停用此Updatable | called by user to disable this Updatable

        Args:

        Returns:
            None
        """
        self._enabled = False

    def enable(self) -> None:
        """
        * 由用户调用，重新启用此Updatable | called by user to re-enable this Updatable

        Args:

        Returns:
            None
        """
        self._enabled = True

    def remove(self) -> None:
        """
        * 由用户调用，删除此Updatable | called by user to remove this Updatable

        Args:

        Returns:
            None
        """
        self._alive = False
        self._enabled = False
        self.halt(self.know)

    @property
    def enabled(self) -> bool:
        """
        * 检查此对象是否在运行 | called by user to check if this object is running
        """
        return self._enabled

    # ==================================================================================

    def handle(self, update_now: bool = False) -> bool:
        """
        * 由用户调用，执行此函数 | called by user to update
        Args:
            update_now: bool  # 是否更新tick | whether Update Tick.

        Returns:
            bool # 是否成功执行更新逻辑 | whether executed init/step logic successfully
        """
        # 更新时间 | Update Tick
        if update_now:
            self.know.now = get.ticks()

        # 检查是否暂停 | Check if pause
        if not self._enabled:
            return False

        # 执行更新逻辑 | exec Update=
        if not self._inited:
            try:
                self.init(self.know)
                self._inited = True
            except Exception as e:
                std.error(f"{self.__kind__}.init", e)

        try:
            self.step(self.know)
        except Exception as e:
            std.error(f"{self.__kind__}.step", e)

        return True


class CreepLogic(Updatable):
    __kind__ = "creep-logic"
    __creep_table__ = {}  # 用于记录CreepLogic的实例 | Used to record instances of CreepLogic

    def __init__(self, creep: Creep = None, spawn: StructureSpawn = None):
        self.creep = creep
        self.spawn = spawn
        self._next = None
        self._next_ref = None
        self.name, self.recipe = self.craft()
        if not creep and not self.recipe:
            std.error("CreepLogic", "CreepLogic must __init__ with a creep or has recipe.")
        if self.name in CreepLogic.__creep_table__:
            std.error("CreepLogic", "CreepLogic name must be unique. Repeat name: " + self.name)
        CreepLogic.__creep_table__[self.name] = self
        super().__init__(False)

    def craft(self) -> tuple[str, list[str]]:
        """
        * 用于生成CreepRecipe | Used to generate CreepRecipe
        Returns:
            tuple[str, list[str]]  # (Creep名称, CreepRecipe) | (Creep name, CreepRecipe)
        """
        return "creep", [MOVE]

    def link(self, next_logic: str):
        """
        * 设置下一个CreepLogic | Set the next CreepLogic
        Args:
            next_logic: str

        Returns:
            None
        """
        self._next = next_logic

    def _enable_next(self):
        ref = self.ref(self._next)
        if not ref: return
        print("enable next", self._next, ref)
        self._next_ref = ref
        ref.enable(self.spawn)

    @classmethod
    def ref(cls, name: str):
        """
        * 通过名称获取CreepLogic实例 | Get CreepLogic instance by name
        Args:
            name: str  # CreepLogic名称 | CreepLogic name

        Returns:
            CreepLogic  # CreepLogic实例 | CreepLogic instance
        """
        if st.na(name): return None
        return cls.__creep_table__.py_get(name)


    def loading(self):
        pass

    def init(self, c: Creep, k: UsrObject, ref:Updatable=None) -> None:
        """
        * 首次执行step函数前，执行此函数 | called before executing the step function for the first time

        Args:
            c: Creep  # 用于记录重要全局数据 | used to record important global data
            k: UsrObject  # 用于记录重要全局数据 | used to record important global data
            ref=None: CreepLogic  # 用于传入下一个CreepLogic | used to pass the next CreepLogic

        Returns:
            None
        """
        pass

    def step(self, c: Creep, k: UsrObject, ref:Updatable=None) -> None:
        """
        * 由handle函数调用，执行此函数 | called by the handle function to execute this function

        Args:
            c: Creep  # 用于记录重要全局数据 | used to record important global data
            k: UsrObject # 用于记录重要全局数据 | used to record important global data
            ref=None: CreepLogic  # 用于传入下一个CreepLogic | used to pass the next CreepLogic

        Returns:
            None
        """
        pass

    def halt(self, c: Creep, k: UsrObject, ref:Updatable=None) -> None:
        """
        * 由外部停止此对象时调用此 | called when this object is stopped externally

        Args:
            c: Creep  # 用于记录重要全局数据 | used to record important global data
            k: UsrObject  # 用于记录重要全局数据 | used to record important global data
            ref=None: CreepLogic  # 用于传入下一个CreepLogic | used to pass the next CreepLogic

        Returns:
            None
        """
        pass

    def enable(self, spawn: StructureSpawn = None):
        if not spawn and not self.spawn:
            std.error("CreepLogic", "CreepLogic __init__ without spawn must enable with a spawn .")
        elif spawn:
            self.spawn = spawn
        print("enable target: ", self.name)
        super().enable()

    def remove(self):
        self._alive = False
        self._enabled = False
        self.halt(self.creep, self.know, self.ref(self._next))

    def handle(self, update_now: bool = False) -> bool:
        """
        * 由用户调用，执行此函数 | called by user to update
        Args:
            update_now: bool  # 是否更新tick | whether Update Tick.

        Returns:
            bool # 是否成功执行更新逻辑 | whether executed init/step logic successfully
        """
        # 更新时间 | Update Tick
        if update_now:
            self.know.now = get.ticks()

        # 检查是否暂停 | Check if pause
        if not self._enabled:
            return False

        # 检查Creep状态
        if not st.creep(self.creep):
            if not self.recipe:  # 表明这是一个一次性的CreepLogic | This is a one-time CreepLogic
                self.remove()
                return False
            elif self.spawn:  # 尝试创建新的Creep | Try to create a new Creep
                self.creep = put.create(self.spawn, self.recipe, self.name)
                if not st.creep(self.creep):
                    return False
            else:
                std.error("CreepLogic", "CreepLogic must __init__ with a creep or has recipe and a spawn.")

        # 执行更新逻辑 | exec Update=
        if not self._inited:
            try:
                self.init(self.creep, self.know, self._next_ref)
                self._inited = True
            except Exception as e:
                std.error(f"{self.__kind__}.init", e)

        # next
        try:
            if st.na(self._next_ref) and self._next:
                self._enable_next()
        except Exception as e:
            std.error(f"{self.__kind__}.enable_next", e)

        # step
        try:
            self.step(self.creep, self.know, self._next_ref)
        except Exception as e:
            std.error(f"{self.__kind__}.step", e)

        return True
