// > sort 6
import {Updatable} from "../build/main.mjs";

class _Engine_Scheduler {
    static instance = null;
    constructor() {
        if (_Engine_Scheduler.instance !== null) {
            return _Engine_Scheduler.instance;
        }
        this.updates = [];
        var _length = len(Updatable.UPDATES);
        for (var i = 0; i < _length; i++) {
            this.updates.push(Updatable.UPDATES[i]());
        }

        // instance
        _Engine_Scheduler.instance = this;
    }

    static remove(name) {
        if (_Engine_Scheduler.instance === null) return;
        var _length = len(_Engine_Scheduler.instance.updates);
        for (var i = 0; i < _length; i++) {
            var item = _Engine_Scheduler.instance.updates[i];
            if (item.name === name) {
                item.remove();
                _Engine_Scheduler.instance.updates.splice(i, 1);
                return;
            }
        }
    }

    static add(obj) {
        if (_Engine_Scheduler.instance === null) return;
        // can not repeat
        if (obj.name){
            var _length = len(_Engine_Scheduler.instance.updates);
            for (var i = 0; i < _length; i++) {
                var item = _Engine_Scheduler.instance.updates[i];
                if (item.name === obj.name) {
                    std.error("Scheduler.add", `name:${obj.name} already exists`);
                }
            }
        }
        _Engine_Scheduler.instance.updates.push(obj);
    }

    static find(name) {
        if (_Engine_Scheduler.instance === null) return null;
        var _length = len(_Engine_Scheduler.instance.updates);
        for (var i = 0; i < _length; i++) {
            var item = _Engine_Scheduler.instance.updates[i];
            if (item.name === name) {
                return item;
            }
        }
        return null;
    }

    get know() {
        return Updatable.KNOW;
    }

    get now() {
        return Updatable.KNOW.now;
    }

    handle() {
        var _length = len(this.updates);
        if (_length > 0) {
            var removes = null;
            for (var i = 0; i < _length; i++) {
                var item = this.updates[i];
                item.handle();
                if (!item._alive){
                    item.remove();
                    if (removes === null) removes = [];
                    removes.push(i);
                }
            }
            if (removes !== null) {
                for (var i = len(removes) - 1; i >= 0; i--) {
                    this.updates.splice(removes[i], 1);
                }
            }
        }
    }
}

function Scheduler() {
    // ADD CLASSES
    /// $UPDATABLE.UPDATES$ ///
    return new _Engine_Scheduler();
}

