
// > sort 3
export var get = {
    ticks: getTicks,
    cpu: getCpuTime,
    heap: getHeapStatistics,
    byid: getObjectById,
    cpu_us: function () {
        return Math.floor(getCpuTime() / 1000);
    },
    get arena() {
        return function () {
            var tmp = UsrObject();
            tmp.py_name = ARENA_NAME;
            tmp.level = ARENA_LEVEL;
            tmp.season = ARENA_SEASON;
            tmp.ticksLimit = ARENA_TICKS_LIMIT;
            tmp.cpuTimeLimit = ARENA_CPU_TIME_LIMIT;
            tmp.cpuTimeLimitFirstTick = ARENA_CPU_TIME_LIMIT_FIRST_TICK;
            return tmp;
        };
    },
    distance: function (obj, target) {
		if (target === null || target === undefined) {
			std.warn('get.distance', 'target is NA. Return 65535');
			return 65535;
		}
		if (obj === null || obj === undefined) {
			std.warn('get.distance', 'obj is NA. Return 65535');
			return 65535;
		}
		if (obj.x === undefined || obj.y === undefined || target.x === undefined || target.y === undefined) {
			std.warn('get.distance', 'obj or target is not a point. Return 65535');
			return 65535;
		}
		return Math.max(Math.abs(obj.x - target.x), Math.abs(obj.y - target.y));
    },
	path: function (obj, target, options=null) {
		if (target === null || target === undefined) {
			std.warn('get.path', 'target is NA. Return null');
			return null;
		}
		if (obj === null || obj === undefined) {
			std.warn('get.path', 'obj is NA. Return null');
			return null;
		}
		var path_res = searchPath (obj, target, options);
		if (path_res.incomplete) {
			std.warn ('get.path', 'Path to target:', target.id, ' is incomplete.');
		}
		return path_res.path;
	},
    find: function(objs, filter_fn) {
		if (len(objs) === 0) {
			return null;
		}
		if (filter_fn === undefined || filter_fn === null) {
			std.warn('get.find', 'filter_fn is NA. Return first object');
		} else {
			if (isinstance(filter_fn, tuple([list, tuple, set]))) {
				filter_fn = std.combo(filter_fn);
			}
			objs = filter(filter_fn, objs);
		}
		return len(objs)?objs[0]:null;
    },
    filter: function(objs, filter_fn) {
		if (len(objs) === 0) {
			return objs;
		}
		if (filter_fn === undefined || filter_fn === null) {
			std.warn('get.find', 'filter_fn is NA. Return the original array|list.');
		} else {
			if (isinstance(filter_fn, tuple([list, tuple, set]))) {
				filter_fn = std.combo(filter_fn);
			}
			return filter(filter_fn, objs);
		}
		return objs;
	},
    classname: function (game_object) {
		if (game_object instanceof Creep) return 'Creep';
		if (game_object instanceof StructureSpawn) return 'Spawn';
		if (game_object instanceof StructureExtension) return 'Extension';
		if (game_object instanceof StructureRoad) return 'Road';
		if (game_object instanceof StructureWall) return 'Wall';
		if (game_object instanceof StructureRampart) return 'Rampart';
		if (game_object instanceof StructureTower) return 'Tower';
		if (game_object instanceof Structure) return 'Structure';
		if (game_object instanceof Source) return 'Source';
		if (game_object instanceof Flag) return 'Flag';
		if (game_object instanceof ConstructionSite) return 'Site';
		if (game_object instanceof Resource) return 'Resource';
		if (game_object instanceof GameObject) return 'GameObject';
		return 'Object';
    },
    terrain: function (x_or_point, y) {
		if (x_or_point === null || x_or_point === undefined) {
			std.warn('get.terrain', 'x_or_point is NA. Return null');
			return null;
		}
		if (y === null || y === undefined) {
			if (x_or_point.x === undefined || x_or_point.y === undefined) {
				std.warn('get.terrain', 'x_or_point is not a point. Return null');
				return null;
			}

			return getTerrainAt(x_or_point);
		}
		return getTerrainAt(Point(x_or_point, y));
    },
    all: function (p_type, filter_fn=null) {
		var objs = getObjectsByPrototype(p_type);
		if (p_type === Creep) {
			var __left0__ = tuple([[], get.ticks()]);
			var tmp = __left0__ [0];
			var now_ticks = __left0__ [1];
			for (var obj of objs) {
				if (!(obj.birth_date) || obj.my && obj.birth_date <= now_ticks) {
					tmp.append(obj);
				}
			}
			objs = tmp;
		}
		if (filter_fn === undefined || filter_fn === null) {
			return objs;
		} else {
			if (isinstance(filter_fn, tuple([list, tuple, set]))) {
				filter_fn = std.combo(filter_fn);
			}
			return filter(filter_fn, objs);
		}
    },
    one: function(p_type, filter_fn=null) {
		var objs = get.all(p_type, filter_fn);
		if (len(objs) === 0) {
			return null;
		}
		return objs[0];
    },
    creeps: function (filter_fn=null) {
		return get.all(Creep, filter_fn);
    },
    creep: function (filter_fn=null) {
		return get.one(Creep, filter_fn);
    },
    friends:function (filter_fn=null) {
		return get.all(Creep, function (obj) {
			return obj.my && (!(filter_fn) || filter_fn(obj));
		});
    },
    friend:function (filter_fn=null) {
		return get.one(Creep, function(obj) {
			return obj.my && (!(filter_fn) || filter_fn(obj));
		});
    },
    enemies:function (filter_fn=null) {
		return get.all(Creep, function (obj) {
			return !(obj.my) && (!(filter_fn) || filter_fn(obj));
		});
    },
    enemy:function (filter_fn=null) {
		return get.one(Creep, function(obj) {
			return !(obj.my) && (!(filter_fn) || filter_fn(obj));
		});
    },
    spawns:function (filter_fn=null) {
		return get.all(StructureSpawn, filter_fn);
    },
    spawn:function (filter_fn=null) {
		return get.one(StructureSpawn, filter_fn);
    },
    structures:function (filter_fn=null) {
		return get.all(Structure, filter_fn);
    },
    structure:function (filter_fn=null) {
		return get.one(Structure, filter_fn);
    },
    sources:function (filter_fn=null) {
		return get.all(Source, filter_fn);
    },
    source:function (filter_fn=null) {
		return get.one(Source, filter_fn);
    },
    sites:function (filter_fn=null) {
		return get.all(ConstructionSite, filter_fn);
    },
    site:function (filter_fn=null) {
		return get.one(ConstructionSite, filter_fn);
    },
    resources:function (filter_fn=null) {
		return get.all(Resource, filter_fn);
    },
    resource:function (filter_fn=null) {
		return get.one(Resource, filter_fn);
    },
    towers:function (filter_fn=null) {
		return get.all(StructureTower, filter_fn);
    },
    tower:function (filter_fn=null) {
		return get.one(StructureTower, filter_fn);
    },
    walls:function (filter_fn=null) {
		return get.all(StructureWall, filter_fn);
    },
    wall:function (filter_fn=null) {
		return get.one(StructureWall, filter_fn);
    },
    ramparts:function (filter_fn=null) {
		return get.all(StructureRampart, filter_fn);
    },
    rampart:function (filter_fn=null) {
		return get.one(StructureRampart, filter_fn);
    },
    extensions:function (filter_fn=null) {
		return get.all(StructureExtension, filter_fn);
    },
    extension:function (filter_fn=null) {
		return get.one(StructureExtension, filter_fn);
    },
    boxes:function (filter_fn=null) {
		return get.all(StructureContainer, filter_fn);
    },
    box:function (filter_fn=null) {
		return get.one(StructureContainer, filter_fn);
    },
    roads:function (filter_fn=null) {
		return get.all(StructureRoad, filter_fn);
    },
    road:function (filter_fn=null) {
		return get.one(StructureRoad, filter_fn);
    },
    flags:function (filter_fn=null) {
		return get.all(Flag, filter_fn);
    },
    flag:function (filter_fn=null) {
		return get.one(Flag, filter_fn);
    },
    score_controllers:function (filter_fn=null) {
		return get.all(ScoreController, filter_fn);
    },
    score_controller:function (filter_fn=null) {
		return get.one(ScoreController, filter_fn);
    },
    storages:function (filter_fn=null) {
		return get.all(GameObject, function (obj) {
			return st.storable(obj) && filter_fn(obj);
		});
    },
    storage:function (filter_fn=null) {
		return get.one(GameObject, function (obj) {
			return st.storable(obj) && filter_fn(obj);
		});
    },
    moved:function (unit, ticks_offset=0) {
		return unit.last_move === (get.ticks() + ticks_offset);
    },
    attacked:function (unit, ticks_offset=0) {
		return unit.last_attack === (get.ticks() + ticks_offset) || unit.last_ranged === (get.ticks() + ticks_offset);
    },
    meleed:function (unit, ticks_offset=0) {
		return unit.last_melee === (get.ticks() + ticks_offset);
    },
    ranged:function (unit, ticks_offset=0) {
		return unit.last_ranged === (get.ticks() + ticks_offset);
    },
    healed:function (unit, ticks_offset=0) {
		return unit.last_heal === (get.ticks() + ticks_offset);
    },
    fetched:function (unit, ticks_offset=0) {
		return unit.last_fetch === (get.ticks() + ticks_offset);
    },
    deposited:function (unit, ticks_offset=0) {
		return unit.last_deposit === (get.ticks() + ticks_offset);
    },
    built:function (unit, ticks_offset=0) {
		return unit.last_build === (get.ticks() + ticks_offset);
    },
    intermited:function (unit, ticks_offset=0) {
		return unit.last_intermit === (get.ticks() + ticks_offset);
    },
    closest:function (obj, objs, filter_fn=null) {
		if (len(objs) === 0) {
			return null;
		}
		if (filter_fn !== undefined && filter_fn !== null) {
			if (isinstance(filter_fn, tuple([list, tuple, set]))) {
				filter_fn = std.combo(filter_fn);
			}
			objs = filter(filter_fn, objs);
		}
		if (len(objs) === 0) {
			std.warn('get.closest', 'No objects left after filtering.')
			return null;
		}
		var ret = findClosestByRange(obj, objs);
		if (ret === undefined) {
			return null;
		}
		return ret;
    },
    quickest:function (obj, objs, filter_fn=null, options=null) {
		if (len(objs) === 0) {
			return null;
		}
		if (filter_fn !== undefined && filter_fn !== null) {
			if (isinstance(filter_fn, tuple([list, tuple, set]))) {
				filter_fn = std.combo(filter_fn);
			}
			objs = filter(filter_fn, objs);
		}
		if (len(objs) === 0) {
			std.warn('get.closest', 'No objects left after filtering.')
			return null;
		}
		var ret = findClosestByPath(obj, objs, options);
		if (ret === undefined) {
			return null;
		}
		return ret;
    },
    inrange:function (obj, objs, range, filter_fn=null) {
		if (len(objs) === 0) {
			return objs;
		}
		if (filter_fn !== undefined && filter_fn !== null) {
			if (isinstance(filter_fn, tuple([list, tuple, set]))) {
				filter_fn = std.combo(filter_fn);
			}
			objs = filter(filter_fn, objs);
		}
		if (len(objs) === 0) {
			std.warn('get.closest', 'No objects left after filtering.')
			return null;
		}
		var tmp = [];
		for (var _obj of objs) {
			if (get.distance(obj, _obj) <= range) {
				tmp.append(_obj);
			}
		}
		return tmp;
    },
    energy:function (target, percent=false) {
		if (target === null || !(target.store)) {
			return -(1);
		}
		var current = target.store.getUsedCapacity(RESOURCE_ENERGY);
		if (current === null) {
			return -(1);
		}
		var value = current;
		if (percent) {
			var total = target.store.getCapacity(RESOURCE_ENERGY);
			if (total === null) {
				return -(1);
			}
			value = int(current / total * 100 + 0.5)
		}
		return value;
    },
    health:function (target, percent=false) {
		if (target === null) {
			return 0;
		}
		if (target.hits === undefined) {
			return 0;
		}
		var current = target.hits;
		var value = current;
		if (percent) {
			if (target.hitsMax === undefined) {
				return 0;
			}
			var total = target.hitsMax;
			value = int(current / total * 100 + 0.5)
		}
		return value;
    },
    parts:function (creep, broken=true) {
		if (!(creep) || !(creep.body)) {
			return [];
		}
		var prts = [];
		for (var prt of creep.body) {
			if (!broken && prt.hits <= 0) continue;
			prts.append(prt.type);
		}
		return prts;
    },
    pcount:function (creep, part_type, broken=true) {
		if (!(creep) || !(creep.body)) {
			return -(1);
		}
		var count = 0;
		for (var part of creep.body) {
			if (part.type === part_type) {
				if (broken || part.hits > 0) {
					count++;
				}
			}
		}
		return count;
    },
    pfirst:function (creep) {
		if (!(creep) || !(creep.body)) {
			return null;
		}
		for (var part of creep.body) {
			if (part.hits > 0) {
				return part;
			}
		}
		return null;
    },
    plast:function (creep) {
		if (!(creep) || !(creep.body)) {
			return null;
		}
		for (var part of py_reversed(creep.body)) {
			if (part.hits <= 0) {
				return part;
			}
		}
		return null;
    },
    wait:function (creep) {
		var move_count = get.pcount(creep, MOVE, false);
		if (move_count === -1) {
			return -1;
		}
		if (move_count === 0) {
			return 65535;
		}
		var fatigue = creep.fatigue;
		if (fatigue <= 0) {
			return 0;
		}
		return Math.ceil((fatigue / 2) / move_count);
    },
    offset:function (pos, dx, dy) {
		return Point(pos.x + dx, pos.y + dy);
    },
    SCORE_TABLE: dict([[ATTACK, 10], [RANGED_ATTACK, 15], [HEAL, 15], [TOUGH, 1], [MOVE, 2], [CARRY, 2], [WORK, 5]]),
    score:function (creep, dhp=null) {
		if (st.jnumber(dhp) && dhp > 0) {
			dhp = math.floor(dhp);
			var bk = get.plast(creep);
			var fr = get.pfirst(creep);
			if (get.health(creep, true) === 100) {
				return 0;
			}
			var fs = int(get.SCORE_TABLE [fr.py_metatype] * (1 - fr.hits / 100));
			if (!(bk)) {
				return fs;
			}
			var bs = int(get.SCORE_TABLE [bk.py_metatype] * (1 - ((dhp - fr.hits) + 100) / 100));
			if (dhp > 100 - fr.hits) {
				return fs + bs;
			}
			return fs;
		} else {
			var __left0__ = tuple([0, 0, 0]);
			var carries = __left0__ [0];
			var moves = __left0__ [1];
			var score = __left0__ [2];
			var prts = get.parts(creep);
			for (var [i, prt] of enumerate(prts)) {
				if (prt === ATTACK) {
					score += 10 + i / 5;
				} else if (prt === RANGED_ATTACK) {
					score += 15 + i / 4;
				} else if (prt === HEAL) {
					score += 15 + i;
				} else {
					score++;
					if (prt === MOVE) {
						moves++;
					} else if (prt === CARRY) {
						carries++;
					}
				}
			}
			var length = (len(prts) - carries) - moves;
			var movements = moves * 2;
			var move_cost = length * 10;
			if (length === 0) {
				var coef = 0;
			} else if (movements > move_cost) {
				var coef = 2;
			} else {
				var coef = (2 * movements) / move_cost;
			}
			var coef = (1 + coef) / 2;
			return math.floor(score * coef);
		}
    }
}