
// > sort 0
export var math = {
    pi: Math.PI,
    e: 2.7182818,
    ln2: Math.LN2,
    ln10: Math.LN10,
    log2e: Math.LOG2E,
    log10e: Math.LOG10E,
    sqrt1_2: Math.SQRT1_2,
    sqrt2: Math.SQRT2,
    sqrt: Math.sqrt,
    pow: Math.pow,
    cbrt: Math.cbrt,
    clamp: function (v, mi, mx) {
        return Math.min(Math.max(v, mi), mx);
    },
    sign: function (x) {
        return x > 0 ? 1 : x < 0 ? -1 : 0;
    },
    sin: Math.sin,
    cos: Math.cos,
    tan: Math.tan,
    cot: function (x) {
        var _tan = Math.tan(x);
        return _tan ? 1 / _tan : Infinity;
    },
    asin: Math.asin,
    acos: Math.acos,
    atan: Math.atan,
    atan2: Math.atan2,
    acot: function (x) {
        return x ? Math.atan(1 / x) : Math.PI / 2;
    },
    degree: function (r) {
        return r * 180 / Math.PI;
    },
    radian: function (d) {
        return d * Math.PI / 180;
    },
    min: Math.min,
    max: Math.max,
    abs: Math.abs,
    floor: Math.floor,
    ceil: Math.ceil,
    round: Math.round,
    trunc: Math.trunc,
    log: Math.log,
    exp: Math.exp,
    log10: Math.log10,
    log2: Math.log2,
    epsilon: function (a, b) {
        return Math.abs(a - b) < 0.000001;
    },
    random: function (n = 1.0) {
        return Math.random() * n;
    },
    irandom: function (n) {
        return Math.floor(Math.random() * (n + 1));
    },
    randint: function (min, max) {
        return Math.floor(Math.random() * (max - min + 1)) + min;
    },
    randfloat: function (min, max) {
        return Math.random() * (max - min) + min;
    },
    randrange: function (min, max, step = undefined) {
        if (step === undefined || step === null) return math.randfloat(min, max);
        if (min > max) [min, max] = [max, min];
        const range = ((max - min) / step) + 1;
        return Math.random() * range * step + min;
    },
    irandrange: function (min, max, step = undefined) {
        return Math.floor(math.randrange(min, max, step));
    },
    choice: function (arr) {
        var _length = len(arr);
        if (_length === 0) std.error("math.choice", "Array is empty.");
        return arr[Math.floor(Math.random() * _length)];
    },
    sample: function (arr, n = 1) {
        var _length = len(arr);
        if (n > _length) std.error("math.sample", `Sample size:${n} is larger than the array-size:${_length}.`);
        let reservoir = copy(arr);
        for (let i = n; i < arr.length; i++) {
            let j = Math.floor(Math.random() * (i + 1));
            if (j < n) {
                reservoir[j] = arr[i];
            }
        }
        return reservoir;
    },
    shuffle: function (arr) {
        var m = len(arr), t, i;
        while (m) {
            i = Math.floor(Math.random() * m--);
            t = arr[m];
            arr[m] = arr[i];
            arr[i] = t;
        }
        return arr;
    }
};