// > sort 4
export var put = {
	_spawn_memory: [],
	__check_unit: function (method_name, unit, method=null) {
		if (!unit) {
            std.warn ('put.' + method_name, 'unit is undefined. Ignore.')
            return false;
        }
        if (!unit.exists) {
			std.warn ('put.' + method_name, `unit:${unit.id} does not exist. Ignore.`)
            return false;
		}
		if (method && !unit[method]) {
			std.warn ('put.' + method_name, `unit:${unit.id} do not have '${method}' method. Ignore.`)
			return false;
		}
		return true;
    },
    __check_target: function (method_name, target, method=null) {
		if (!target) {
            std.warn ('put.' + method_name, 'target is undefined. Ignore.')
            return false;
        }
        if (!target.exists) {
			std.warn ('put.' + method_name, `target:${target.id} does not exist. Ignore.`)
            return false;
		}
		if (method && !target[method]) {
			std.warn ('put.' + method_name, `target:${target.id} do not have '${method}' method. Ignore.`)
			return false;
		}
		return true;
    },
	move:function (unit, to, options=null) {
		if (!put.__check_unit('put.move', unit, 'moveTo')) return -1;
		if (to === null || to === undefined) {
			std.warn('put.move', `unit:${unit.id} -> target:NA. Ignore.`)
			return -1;
		}
		if (to.x !== undefined && to.y !== undefined) {
			if (!(unit.path_goal) || unit.path_goal.x !== to.x && unit.path_goal.y !== to.y) {
				var path_res = searchPath (unit, to, options);
				if (path_res.incomplete) {
					std.warn ('put.move', 'Path to target:', to, ' is incomplete.');
				}
				unit.path = path_res.path;
				unit.path_goal = to;
				unit.path_index = 0;
			}
			if (unit.path) {
				to = unit.path [unit.path_index];
				unit.path_index++;
				if (unit.path_index >= unit.path.length) {
					unit.path = null;
					unit.path_index = 0;
				}
			}
			var _ = unit.moveTo (to);
		}
		else {
			var _ = unit.move (to);
		}
		if (_ === OK) {
			unit.last_move = get.ticks ();
		}
		return _;
	},
	push:function (unit, target, options=null) {
		if (!put.__check_unit('put.push', unit, 'moveTo')) return -1;
		if (!put.__check_target('put.push', target, 'pull')) return ERR_INVALID_TARGET;
		var _ = put.move (unit, target, options);
		if (_ !== OK) {
			return _;
		}
		return target.pull (unit);
	},
	attack:function (unit, target, move=true) {
		if (!put.__check_unit('put.attack', unit, 'attack')) return -1;
		if (!put.__check_target('put.attack', target)) return ERR_INVALID_TARGET;
		if (unit instanceof StructureTower) return {attack: unit.attack(target)};
		var melee = st.melee (unit);
		var ranged = st.ranged (unit);
		var tick = get.ticks ();
		var dist = get.distance (unit, target);
		var result = {};
		if (dist === 1) {
			if (melee) {
				result.melee = unit.attack (target);
			}
			if (ranged) {
				result.ranged = unit.rangedMassAttack ();
			}
			if (move && !melee && st.atkable (target)) {
				result.move = put.escape (unit, target, (move === true ? null : move));
			}
		}
		else if (dist <= 3) {
			if (ranged) {
				result.ranged = unit.rangedAttack (target);
				if (move && dist <= 2) {
					result.move = put.escape (unit, target, (move === true ? null : move));
				}
			}
			else if (move && melee) {
				result.move = put.move (unit, target, (move === true ? null : move));
			}
		}
		else if (move) {
			result.move = put.move (unit, target, (move === true ? null : move));
		}
		if (result.melee === OK) {
			unit.last_melee = tick;
		}
		if (result.ranged === OK) {
			unit.last_ranged = tick;
		}
		return result;
	},
	heal:function (unit, target, move=true) {
		if (!put.__check_unit('put.heal', unit, 'heal')) return -1;
		if (!put.__check_target('put.heal', target)) return ERR_INVALID_TARGET;
		if (unit instanceof StructureTower) return {heal: unit.heal(target)};
		var result = {};
		var tick = get.ticks ();
		var dist = get.distance (unit, target);
		if (dist <= 1) {
			if (!(unit.last_melee) || unit.last_melee !== tick) {
				result.heal = unit.heal (target);
			}
			else if (!(unit.last_ranged) || unit.last_ranged !== tick) {
				result.ranged = unit.rangedHeal (target);
			}
		}
		else if (dist <= 3) {
			result.ranged = unit.rangedHeal (target);
		}
		else if (move) {
			result.move = put.move (unit, target, (move === true ? null : move));
		}
		if (result.heal === OK || result.ranged === OK) {
			unit.last_heal = tick;
		}
		return result;
	},
	fetch:function (unit, target, resource_type=null, amount=null, move=true) {
		if (!put.__check_unit('put.fetch', unit)) return -1;
		if (!put.__check_target('put.fetch', target)) return ERR_INVALID_TARGET;
		if (resource_type === null || resource_type === undefined) resource_type = RESOURCE_ENERGY;
		var tick = get.ticks ();
		if (move && get.distance (unit, target) > 1) {
			return put.move (unit, target, (move === true ? null : move));
		}
		if((target instanceof Resource) && target.resourceType === RESOURCE_ENERGY) return unit.pickup(target);
		unit.last_fetch = tick;
		if (amount) {
			if(target instanceof Creep) return target.transfer(unit, resource_type, amount);
			if(target instanceof Structure) return unit.withdraw(target, resource_type, amount);
		}
		else {
			if(target instanceof Creep) return target.transfer(unit, resource_type);
			if(target instanceof Structure) return unit.withdraw(target, resource_type);
		}
		return ERR_INVALID_TARGET;
	},
	deposit:function (unit, target, resource_type=null, amount=null, move=true) {
		if (!put.__check_unit('put.deposit', unit)) return -1;
		// if (!put.__check_target('put.deposit', target)) return ERR_INVALID_TARGET;
		// target can be null
		if (resource_type === null || resource_type === undefined) resource_type = RESOURCE_ENERGY;
		unit.last_deposit = get.ticks();
		if (target === null) {
			return unit.drop (resource_type, amount);
		}
		if (move && get.distance (unit, target) > 1) {
			return put.move (unit, target, (move === true ? null : move));
		}
		if (amount) {
			if (st.storable (target)) {
				return unit.transfer (target, resource_type, amount);
			}
			return unit.drop (resource_type, amount);
		}
		else {
			if (st.storable (target)) {
				return unit.transfer (target, resource_type);
			}
			return unit.drop (resource_type);
		}
	},
	build:function (unit, site, move=true) {
		if (!put.__check_unit('put.build', unit, "build")) return -1;
		if (!put.__check_target('put.build', site)) return ERR_INVALID_TARGET;
		unit.last_build = get.ticks();
		if (move && get.distance (unit, site) > 3) {
			return put.move (unit, site, (move === true ? null : move));
		}
		return unit.build (site);
	},
	harvest:function (unit, target, move=true) {
		if (!put.__check_unit('put.harvest', unit, "harvest")) return -1;
		if (!put.__check_target('put.harvest', target)) return ERR_INVALID_TARGET;
		if (!st.source(target)){
			std.warn('put.harvest', `target:${target.id} is not a source. Ignore.`)
			return ERR_INVALID_TARGET;
		}
		unit.last_harvest = get.ticks();
		if (move && get.distance (unit, target) > 1) {
			return put.move (unit, target, (move === true ? null : move));
		}
		return unit.harvest (target);
	},
	escape:function (unit, target, options=null) {
		if (!put.__check_unit('put.escape', unit)) return -1;
		if (!put.__check_target('put.escape', target)) return ERR_INVALID_TARGET;
		var pos = Point (unit.x + math.sign (unit.x - target.x), unit.y + math.sign (unit.y - target.y));
		if (get.terrain (pos) === TERRAIN_WALL) {
			var spawn = get.spawn (st.friend);
			if (spawn) {
				return put.move (unit, spawn, options);
			}
		}
		return put.move (unit, pos, options);
	},
	intermit:function (unit, target, resource_type=null, options=null) {
		if (!put.__check_unit('put.intermit', unit)) return -1;
		if (!put.__check_target('put.intermit', target)) return ERR_INVALID_TARGET;
		if (resource_type === null || resource_type === undefined) resource_type = RESOURCE_ENERGY;
		var tick = get.ticks ();
		var drops = get.resources ((function __lambda__ (obj) {
			return obj.resourceType === resource_type && unit._last_idrop && unit._last_idrop.x === obj.x && unit._last_idrop.y == obj.y;
		}));
		var free = unit.store.getFreeCapacity (resource_type);
		if (len (drops) && free > 0) {
			unit.last_intermit = get.ticks ();
			return put.fetch (unit, drops [0], resource_type, free, false);
		}
		var current = unit.store.getUsedCapacity (resource_type);
		if (current === 0) {
			return DONE;
		}
		unit.last_intermit = tick;
		var dist = get.distance (unit, target);
		if (dist <= 1) {
			var _ = put.deposit (unit, target, resource_type, null, (options ? options : true));
			if (_ !== OK) {
				return _;
			}
			return DONE;
		}
		else {
			var _ = put.deposit (unit, null, resource_type, null, (options ? options : true));
			if (unit._last_idrop) {
				unit._last_idrop.x = unit.x;
				unit._last_idrop.y = unit.y;
			}
			else {
				unit._last_idrop = Point (unit.x, unit.y);
			}
			if (_ !== OK) {
				return _;
			}
			return put.move (unit, target, options);
		}
	},
	follow:function (unit, target, distance, options=null, move=true) {
		if (!put.__check_unit('put.follow', unit)) return -1;
		if (!put.__check_target('put.follow', target)) return ERR_INVALID_TARGET;
		var target_is_creep = st.creep (target);
		var __left0__ = (isinstance (distance, tuple) ? distance : tuple ([distance, 65535]));
		var stop_dist = __left0__ [0];
		var ignore_dist = __left0__ [1];
		var dist = get.distance (unit, target);
		if (target.my) {
			if (move && target_is_creep && (ignore_dist >= dist && dist >= stop_dist)) {
				var _ = put.move (target, unit, (move === true ? null : move));
				if (_ !== OK) {
					return _;
				}
			}
			return put.move (unit, target, options);
		} else {
			if (dist <= stop_dist) {
				return put.escape (unit, target, options);
			}
			return put.move (unit, target, options);
		}
	},
	carry:function (unit, src, dst, resource_type=null, options=null, intermit=false) {
		if (!put.__check_unit('put.carry', unit)) return -1;
		if (!put.__check_target('put.carry', src)) return ERR_INVALID_TARGET;
		if (resource_type === null || resource_type === undefined) resource_type = RESOURCE_ENERGY;
		if (!(intermit)) {
			if (get.energy (unit) === 0) {
				if (get.distance (unit, src) <= 1) {
					put.fetch (unit, src, resource_type, null, false);
					return put.move (unit, dst, options);
				}
				return put.fetch (unit, src, resource_type, null, (options ? options : true));
			}
			if (get.distance (unit, dst) <= 1) {
				put.deposit (unit, dst, resource_type, null, false);
				return put.move (unit, src, options);
			}
			return put.deposit (unit, dst, resource_type, null, (options ? options : true));
		}
		else {
			var drops = get.resources (function (obj) {
				return obj.resourceType === resource_type && get.distance (unit, obj) <= 1;
			});
			var free = unit.store.getFreeCapacity (resource_type);
			if (len (drops) && free > 0) {
				return put.fetch (unit, drops [0], resource_type, free, false);
			}
			var current = unit.store.getUsedCapacity (resource_type);
			if (current === 0) {
				return put.fetch (unit, src, resource_type, null, (options ? options : true));
			}
			var dist = get.distance (unit, dst);
			if (dist <= 1) {
				var _ = put.deposit (unit, dst, resource_type, null, (options ? options : true));
				if (_ !== OK) {
					return _;
				}
				put.move (unit, src, options);
				return DONE;
			}
			else {
				var _ = put.deposit (unit, null, resource_type, null, (options ? options : true));
				if (_ !== OK) {
					return _;
				}
				return put.move (unit, dst, options);
			}
		}
	},
	site:function (x_pos, y_type, building_type=null, single=false) {
		if (!(building_type)) {
			var building_type = y_type;
			var x_pos = x_pos.x;
			var y_type = x_pos.y;
		}
		if (single) {
			var exists = get.site (function (ste) {
				return ste.x === x_pos && ste.y === y_type && ste.structureType === building_type;
			});
			if (exists) {
				return ERR_NAME_EXISTS;
			}
		}
		var _ = createConstructionSite (x_pos, y_type, building_type);
		if (_.object !== undefined) {
			return _.object;
		}
		return _.error;
	},
	create:function (spawn, recipe, py_name=null) {
		var ready_index = -1;
		var now_ticks = get.ticks ();
		for (var i = 0; i < len (put._spawn_memory); i++) {
			if (put._spawn_memory [i].birth_date <= now_ticks) {
				ready_index = i;
				break;
			}
		}
		if (ready_index !== -1) {
			return put._spawn_memory.py_pop (ready_index);
		}
		if (spawn.lock) {
			if (now_ticks <= spawn.lock) {
				return ERR_BUSY;
			}
			spawn.lock = undefined;
		}
		var info = spawn.spawnCreep (recipe);
		if (info.error) {
			return info.error;
		}
		spawn.lock = now_ticks + 2;
		if (py_name) {
			info.object.py_name = py_name;
		}
		info.object.birth_date = get.ticks () + CREEP_SPAWN_TIME * len (recipe);
		put._spawn_memory.append (info.object);
		return ERR_BUSY;
	}
}
