// > sort 1
export var st = {
    // -------------- JS基础类型判断 | JS Basic Type Judgment -----------------
    NOT_JNUMBER: 'obj is not a js number.',
    NOT_JLIST: 'obj is not a js list.',
    NOT_JDICT: 'obj is not a js dict.',
    NOT_JSTRING: 'obj is not a js string.',
    NOT_JBOOL: 'obj is not a js boolean.',
    // -------------- PY基础类型判断 | PY Basic Type Judgment -----------------

    NOT_LIST: 'obj is not a py list.',
    NOT_TUPLE: 'obj is not a py tuple.',
    NOT_DICT: 'obj is not a py dict.',
    NOT_SET: 'obj is not a py set.',
    NOT_STR: 'obj is not a py str.',
    NOT_INT: 'obj is not a py int.',
    NOT_FLOAT: 'obj is not a py float.',
    NOT_BOOL: 'obj is not a py bool.',
    // -------------- 游戏对象类型判断 | Game Object Type Judgment -----------------

    NOT_SITE: 'obj is not a site.',
    NOT_CREEP: 'obj is not a creep.',
    NOT_GOBJECT: 'obj is not a game object.',
    NOT_OSTRUCTURE: 'obj is not an owned structure.',
    NOT_RESOURCE: 'obj is not a resource.',
    NOT_SOURCE: 'obj is not a source.',
    NOT_STORE: 'obj is not a store.',
    NOT_STRUCTURE: 'obj is not a structure.',
    NOT_BOX: 'obj is not a box.',
    NOT_EXTENSION: 'obj is not an extension.',
    NOT_RAMPART: 'obj is not a rampart.',
    NOT_ROAD: 'obj is not a road.',
    NOT_SPAWN: 'obj is not a spawn.',
    NOT_TOWER: 'obj is not a tower.',
    NOT_WALL: 'obj is not a wall.',
    NOT_SPAWNING: 'obj is not a spawning.',
    NOT_FLAG: 'obj is not a flag.',
    NOT_POINT: 'obj is not a point.',
    // -------------- 对象能力类型判断 | Object Ability Type Judgment -----------------

    NOT_MOVABLE: 'obj is not movable.',
    NOT_ATKABLE: 'obj is not atkable.',
    NOT_MELEE: 'obj is not melee.',
    NOT_RANGED: 'obj is not ranged.',
    NOT_HEALABLE: 'obj is not healable.',
    NOT_HITABLE: 'obj is not hitable.',
    NOT_WORKABLE: 'obj is not workable.',
    NOT_STORABLE: 'obj is not storable.',
    NOT_ENERGETIC: 'obj is not energetic.',
    NOT_DAMAGED: 'obj is not damaged.',
    NOT_EMPTY: 'obj is not empty.',
    NOT_FULL: 'obj is not full.',
    NOT_MINE: 'obj is not mine.',
    NOT_FRIEND: 'obj is not a friend.',
    NOT_ENEMY: 'obj is not an enemy.',
    // -------------- Other -----------------

    POOR: 'Not enough energy or resources.',
    BUSY: 'The structure is busy.',
    INVALID: 'Invalid arguments or target.',
    OUT_RANGE: 'The target is out of range.',
    // -------------------------------------- JS基础类型判断 | JS Basic Type Judgment -----------------
    jnumber: function (obj) {
        return typeof obj === 'number';
    },
    jlist: function (obj) {
        return obj instanceof Array;
    },
    jstring: function (obj) {
        return typeof obj === 'string';
    },
    jbool: function (obj) {
        return typeof obj === 'boolean';
    },
    jdict: function (obj) {
        return obj instanceof Object;
    },
    na: function (obj) {
        return obj === null || obj === undefined;
    },
    // -------------------------------------- PY基础类型判断 | PY Basic Type Judgment -----------------
    list: function (obj) {
        return isinstance(obj, list);
    },
    tuple: function (obj) {
        return isinstance(obj, tuple);
    },
    dict: function (obj) {
        return isinstance(obj, dict);
    },
    set: function (obj) {
        return isinstance(obj, set);
    },
    str: function (obj) {
        return isinstance(obj, str);
    },
    int: function (obj) {
        return isinstance(obj, int);
    },
    float: function (obj) {
        return isinstance(obj, float);
    },
    bool: function (obj) {
        return isinstance(obj, bool);
    },
    // -------------------------------------- 游戏对象类型判断 | Game Object Type Judgment -----------------
    site: function (obj) {
        return obj instanceof ConstructionSite;
    },
    creep: function (obj) {
        return obj instanceof Creep && obj.exists;
    },
    gobject: function (obj) {
        return obj instanceof GameObject && obj.exists;
    },
    ostructure: function (obj) {
        return obj instanceof OwnedStructure && obj.exists;
    },
    resource: function (obj) {
        return obj instanceof Resource && obj.exists;
    },
    source: function (obj) {
        return obj instanceof Source && obj.exists;
    },
    store: function (obj) {
        return obj instanceof Store;
    },
    structure: function (obj) {
        return obj instanceof Structure && obj.exists;
    },
    box: function (obj) {
        return obj instanceof StructureContainer && obj.exists;
    },
    extension: function (obj) {
        return obj instanceof StructureExtension && obj.exists;
    },
    rampart: function (obj) {
        return obj instanceof StructureRampart && obj.exists;
    },
    road: function (obj) {
        return obj instanceof StructureRoad && obj.exists;
    },
    spawn: function (obj) {
        return obj instanceof StructureSpawn && obj.exists;
    },
    tower: function (obj) {
        return obj instanceof StructureTower && obj.exists;
    },
    wall: function (obj) {
        return obj instanceof StructureWall && obj.exists;
    },
    spawning: function (obj) {
        return obj instanceof Spawning;
    },
    flag: function (obj) {
        return obj instanceof Flag;
    },
    // -------------------------------------- 对象能力判断 | Object Ability Type Judgment -----------------
    builtable: function (obj) {
        return obj instanceof ConstructionSite && obj.progress < obj.progressTotal;
    },
    movable: function (obj) {
        if (obj && obj.body) {
            for (var p of obj.body) {
                if (p.type === MOVE) return true;
            }
        }
        return false;
    },
    atkable: function (obj) {
        if (obj) {
            if (obj.body) {
                for (var p of obj.body) {
                    if (p.type === ATTACK || p.type === RANGED_ATTACK) return true;
                }
            } else if (obj.attack) {
                return true;
            }
        }
        return false;
    },
    melee: function (obj) {
        if (obj && obj.body) {
            for (var p of obj.body) {
                if (p.type === ATTACK) return true;
            }
        }
        return false;
    },
    ranged: function (obj) {
        if (obj) {
            if (obj.body) {
                for (var p of obj.body) {
                    if (p.type === RANGED_ATTACK) return true;
                }
            } else if (obj.attack) {
                return true;
            }
        }
        return false;
    },
    healable: function (obj) {
        if (obj) {
            if (obj.body) {
                for (var p of obj.body) {
                    if (p.type === HEAL) return true;
                }
            } else if (obj.heal) {
                return true;
            }
        }
        return false;
    },
    hitable: function (obj) {
        return obj && obj.hits !== undefined;
    },
    workable: function (obj) {
        if (obj && obj.body) {
            for (var p of obj.body) {
                if (p.type === WORK) return true;
            }
        }
        return false;
    },
    storable: function (obj) {
        if (obj instanceof Structure && obj.store) return true;
        if (obj instanceof Creep && obj.body) {
            for (var p of obj.body) {
                if (p.type === CARRY) return true;
            }
        }
        return false;
    },
    energetic: function (obj) {
        if (obj instanceof Structure && obj.store && get.energy(obj) > 0) return true;
        if (obj instanceof Creep && get.energy(obj) > 0) return true;
        if (obj instanceof Resource && obj.amount > 0) return true;
        return false;
    },
    damaged: function (obj) {
        return obj.hits < obj.hitsMax;
    },
    empty: function (obj) {
        if (!obj || !obj.store) {
            return true;
        }
        return get.energy(obj) === 0;
    },
    full: function (obj) {
        if (!obj || !obj.store) {
            return true;
        }
        return get.energy(obj, true, '=100');
    },
    friend: function (obj) {
        return obj && obj.my;
    },
    my: function (obj) {
        return obj && obj.my;
    },
    enemy: function (obj) {
        return obj && !(obj.my);
    },
    point: function (obj) {
        return obj.x !== undefined && obj.y !== undefined;
    },
};
